# Copyright (C) 2020  Christian Ledermann
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA

"""Helper functions and classes."""
import logging
from typing import Any
from typing import Callable
from typing import Optional

from fastkml import config
from fastkml.types import Element

logger = logging.getLogger(__name__)


def o_to_attr(
    obj: object,
    element: Element,
    kml_attr: str,
    obj_attr: str,
    required: bool,
    **kwargs: Any,
) -> None:
    """Set an attribute on an KML Element from an object attribute."""
    attribute = getattr(obj, obj_attr)
    if attribute:
        element.set(kml_attr, str(attribute))
    elif required:
        logger.warning(
            "Required attribute '%s' for '%s' missing.",
            obj_attr,
            obj.__class__.__name__,
        )


def o_from_attr(
    obj: object,
    element: Element,
    kml_attr: str,
    obj_attr: str,
    required: bool,
    **kwargs: Any,
) -> None:
    """Set an attribute on self from an KML attribute."""
    attribute = element.get(kml_attr)
    if attribute:
        setattr(obj, obj_attr, attribute)
    elif required:
        logger.warning(
            "Required attribute '%s' for '%s' missing.",
            kml_attr,
            obj.__class__.__name__,
        )


def o_int_from_attr(
    obj: object,
    element: Element,
    kml_attr: str,
    obj_attr: str,
    required: bool,
    **kwargs: Any,
) -> None:
    """Set an attribute on self from an KML attribute."""
    try:
        attribute = int(element.get(kml_attr))
    except (ValueError, TypeError):
        attribute = None
    if attribute is not None:
        setattr(obj, obj_attr, attribute)
    elif required:
        logger.warning(
            "Required attribute '%s' for '%s' missing.",
            kml_attr,
            obj.__class__.__name__,
        )


def o_from_subelement_text(
    obj: object,
    element: Element,
    kml_attr: str,
    obj_attr: str,
    required: bool,
    validator: Optional[Callable[..., bool]] = None,
    **kwargs: Any,
) -> None:
    """Set an attribute on self from the text of a SubElement."""
    elem = element.find(f"{obj.ns}{kml_attr}")  # type: ignore[attr-defined]
    if elem is not None:
        if validator is not None and not validator(elem.text):
            logger.warning(
                "Invalid value for attribute '%s' for '%s'",
                kml_attr,
                obj.__class__.__name__,
            )
        else:
            setattr(obj, obj_attr, elem.text)
    elif required:  # type: ignore[unreachable]
        logger.warning(
            "Required attribute '%s' for '%s' missing.",
            kml_attr,
            obj.__class__.__name__,
        )


def o_to_subelement_text(
    obj: object,
    element: Element,
    kml_attr: str,
    obj_attr: str,
    required: bool,
    validator: Optional[Callable[..., bool]] = None,
    **kwargs: Any,
) -> None:
    """Set the text of a SubElement from an object attribute."""
    attribute = getattr(obj, obj_attr)
    if attribute:
        if validator is not None and not validator(attribute):
            logger.warning(
                "Invalid value for attribute '%s' for '%s'",
                obj_attr,
                obj.__class__.__name__,
            )
        else:
            elem = config.etree.SubElement(  # type: ignore[attr-defined]
                element,
                f"{obj.ns}{kml_attr}",  # type: ignore[attr-defined]
            )
            elem.text = str(attribute)
    elif required:
        logger.warning(
            "Required attribute '%s' for '%s' missing.",
            obj_attr,
            obj.__class__.__name__,
        )



# Copyright (C) 2021  Christian Ledermann
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA

"""Test the Atom classes."""

from fastkml import atom
from tests.base import Lxml
from tests.base import StdLibrary


class TestStdLibrary(StdLibrary):
    """Test with the standard library."""

    def test_atom_link_ns(self) -> None:
        ns = "{http://www.opengis.net/kml/2.2}"  # noqa: FS003
        l = atom.Link(ns=ns)
        assert l.ns == ns
        assert l.to_string().startswith(
            '<kml:link xmlns:kml="http://www.opengis.net/kml/2.2"'
        )

    def test_atom_link(self) -> None:
        l = atom.Link(
            href="#here",
            rel="alternate",
            type="text/html",
            hreflang="en",
            title="Title",
            length=3456,
        )

        serialized = l.to_string()

        assert '<atom:link xmlns:atom="http://www.w3.org/2005/Atom"' in serialized
        assert 'href="#here"' in serialized
        assert 'rel="alternate"' in serialized
        assert 'type="text/html"' in serialized
        assert 'hreflang="en"' in serialized
        assert 'title="Title"' in serialized
        assert 'length="3456"' in serialized

    def test_atom_link_read(self) -> None:
        l = atom.Link()
        l.from_string(
            '<atom:link xmlns:atom="http://www.w3.org/2005/Atom" '
            'href="#here" rel="alternate" type="text/html" hreflang="en" '
            'title="Title" length="3456" />'
        )
        assert l.href == "#here"
        assert l.rel == "alternate"
        assert l.type == "text/html"
        assert l.hreflang == "en"
        assert l.title == "Title"
        assert l.length == 3456

    def test_atom_link_read_no_href(self) -> None:
        l = atom.Link()
        l.from_string(
            '<atom:link xmlns:atom="http://www.w3.org/2005/Atom" '
            'rel="alternate" type="text/html" hreflang="en" '
            'title="Title" length="3456" />'
        )
        assert l.href is None

    def test_atom_person_ns(self) -> None:
        ns = "{http://www.opengis.net/kml/2.2}"  # noqa: FS003
        p = atom._Person(ns=ns)
        assert p.ns == ns

    def test_atom_author(self) -> None:
        a = atom.Author(
            name="Nobody", uri="http://localhost", email="cl@donotreply.com"
        )

        serialized = a.to_string()
        assert '<atom:author xmlns:atom="http://www.w3.org/2005/Atom">' in serialized
        assert "<atom:name>Nobody</atom:name>" in serialized
        assert "<atom:uri>http://localhost</atom:uri>" in serialized
        assert "<atom:email>cl@donotreply.com</atom:email>" in serialized
        assert "</atom:author>" in serialized

    def test_atom_author_read(self) -> None:
        a = atom.Author()
        a.from_string(
            '<atom:author xmlns:atom="http://www.w3.org/2005/Atom">'
            "<atom:name>Nobody</atom:name><atom:uri>http://localhost</atom:uri>"
            "<atom:email>cl@donotreply.com</atom:email></atom:author>",
        )

        assert a.name == "Nobody"
        assert a.uri == "http://localhost"
        assert a.email == "cl@donotreply.com"

    def test_atom_contributor_read_no_name(self) -> None:
        a = atom.Contributor()
        a.from_string(
            '<atom:contributor xmlns:atom="http://www.w3.org/2005/Atom">'
            "<atom:uri>http://localhost</atom:uri>"
            "<atom:email>cl@donotreply.com</atom:email></atom:contributor>",
        )

        assert a.name is None
        assert a.uri == "http://localhost"
        assert a.email == "cl@donotreply.com"

    def test_atom_contributor_no_name(self) -> None:
        a = atom.Contributor(uri="http://localhost", email="cl@donotreply.com")

        assert a.name is None
        assert "atom:name" not in a.to_string()

    def test_author_roundtrip(self) -> None:
        a = atom.Author(name="Christian Ledermann")
        a.uri = "http://iwlearn.net"
        a.email = "christian@gmail.com"
        a.email = "christian"
        assert "email>" not in str(a.to_string())
        a2 = atom.Author()
        a2.from_string(a.to_string())
        assert a.to_string() == a2.to_string()

    def test_link_roundtrip(self) -> None:
        l = atom.Link(href="http://localhost/", rel="alternate")
        l.title = "Title"
        l.type = "text/html"
        l.hreflang = "en"
        l.length = 4096
        l2 = atom.Link()
        l2.from_string(l.to_string())
        assert l.to_string() == l2.to_string()


class TestLxml(Lxml, TestStdLibrary):
    """Test with lxml."""



# Copyright (C) 2021  Christian Ledermann
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA

"""Base classes to run the tests both with the std library and lxml."""
import xml.etree.ElementTree

import pytest

try:  # pragma: no cover
    import lxml  # noqa: F401

    LXML = True
except ImportError:  # pragma: no cover
    LXML = False

from fastkml import config


class StdLibrary:
    """Configure test to run with the standard library."""

    def setup_method(self) -> None:
        """Always test with the same parser."""
        config.set_etree_implementation(xml.etree.ElementTree)
        config.set_default_namespaces()


@pytest.mark.skipif(not LXML, reason="lxml not installed")
class Lxml:
    """
    Configure test to run with lxml.

    Use this mixin as the first base class in the test classes.
    """

    def setup_method(self) -> None:
        """Always test with the same parser."""
        config.set_etree_implementation(lxml.etree)
        config.set_default_namespaces()



# Copyright (C) 2012 - 2023  Christian Ledermann
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
"""Date and time handling in KML."""
import re
from datetime import date
from datetime import datetime
from typing import Optional
from typing import Union

# note that there are some ISO 8601 timeparsers at pypi
# but in my tests all of them had some errors so we rely on the
# tried and tested dateutil here which is more stable. As a side effect
# we can also parse non ISO compliant dateTimes
import dateutil.parser

import fastkml.config as config
from fastkml.base import _BaseObject
from fastkml.enums import DateTimeResolution
from fastkml.types import Element

# regular expression to parse a gYearMonth string
# year and month may be separated by a dash or not
# year is always 4 digits, month is always 2 digits
year_month = re.compile(r"^(?P<year>\d{4})(?:-?)(?P<month>\d{2})$")


class KmlDateTime:
    """A KML DateTime object.

    This class is used to parse and format KML DateTime objects.

    A KML DateTime object is a string that conforms to the ISO 8601 standard
    for date and time representation. The following formats are supported:

    - yyyy-mm-ddThh:mm:sszzzzzz
    - yyyy-mm-ddThh:mm:ss
    - yyyy-mm-dd
    - yyyy-mm
    - yyyy

    The T is the separator between the date and the time, and the time zone
    is either Z (for UTC) or zzzzzz, which represents ±hh:mm in relation to
    UTC. Additionally, the value can be expressed as a date only.

    The precision of the DateTime is dictated by the DateTime value
    which can be one of the following:

    - dateTime gives second resolution
    - date gives day resolution
    - gYearMonth gives month resolution
    - gYear gives year resolution

    The KmlDateTime class can be used to parse a KML DateTime string into a
    Python datetime object, or to format a Python datetime object into a
    KML DateTime string.

    The KmlDateTime class is used by the TimeStamp and TimeSpan classes.
    """

    def __init__(
        self,
        dt: Union[date, datetime],
        resolution: Optional[DateTimeResolution] = None,
    ):
        """Initialize a KmlDateTime object."""
        self.dt = dt
        self.resolution = resolution
        if resolution is None:
            # sourcery skip: swap-if-expression
            self.resolution = (
                DateTimeResolution.date
                if not isinstance(dt, datetime)
                else DateTimeResolution.datetime
            )

    def __bool__(self) -> bool:
        """Return True if the date or datetime is valid."""
        return isinstance(self.dt, date)

    def __eq__(self, other: object) -> bool:
        """Return True if the two objects are equal."""
        return (
            self.dt == other.dt and self.resolution == other.resolution
            if isinstance(other, KmlDateTime)
            else False
        )

    def __repr__(self) -> str:
        """Return a string representation of the object."""
        return f"{self.__class__.__name__}({self.dt!r}, {self.resolution})"

    def __str__(self) -> str:
        """Return the KML DateTime string representation of the object."""
        if self.resolution == DateTimeResolution.year:
            return self.dt.strftime("%Y")
        if self.resolution == DateTimeResolution.year_month:
            return self.dt.strftime("%Y-%m")
        if self.resolution == DateTimeResolution.date:
            return (
                self.dt.date().isoformat()
                if isinstance(self.dt, datetime)
                else self.dt.isoformat()
            )
        if self.resolution == DateTimeResolution.datetime:
            return self.dt.isoformat()
        raise ValueError

    @classmethod
    def parse(cls, datestr: str) -> Optional["KmlDateTime"]:
        """Parse a KML DateTime string into a KmlDateTime object."""
        resolution = None
        dt = None
        if len(datestr) == 4:
            year = int(datestr)
            dt = datetime(year, 1, 1)
            resolution = DateTimeResolution.year
        elif len(datestr) in {6, 7}:
            ym = year_month.match(datestr)
            if ym:
                year = int(ym.group("year"))
                month = int(ym.group("month"))
                dt = datetime(year, month, 1)
                resolution = DateTimeResolution.year_month
        elif len(datestr) in {8, 10}:  # 8 is YYYYMMDD, 10 is YYYY-MM-DD
            dt = dateutil.parser.parse(datestr)
            resolution = DateTimeResolution.date
        elif len(datestr) > 10:
            dt = dateutil.parser.parse(datestr)
            resolution = DateTimeResolution.datetime
        return cls(dt, resolution) if dt else None


class _TimePrimitive(_BaseObject):
    """
    This is an abstract element and cannot be used directly in a KML file.

    This element is extended by the <TimeSpan> and <TimeStamp> elements.
    https://developers.google.com/kml/documentation/kmlreference#timeprimitive
    """


class TimeStamp(_TimePrimitive):
    """Represents a single moment in time."""

    __name__ = "TimeStamp"

    def __init__(
        self,
        ns: Optional[str] = None,
        id: Optional[str] = None,
        target_id: Optional[str] = None,
        timestamp: Optional[KmlDateTime] = None,
    ) -> None:
        super().__init__(ns=ns, id=id, target_id=target_id)
        self.timestamp = timestamp

    def etree_element(self) -> Element:
        element = super().etree_element()
        when = config.etree.SubElement(  # type: ignore[attr-defined]
            element, f"{self.ns}when"
        )
        when.text = str(self.timestamp)
        return element

    def from_element(self, element: Element) -> None:
        super().from_element(element)
        when = element.find(f"{self.ns}when")
        if when is not None:
            self.timestamp = KmlDateTime.parse(when.text)


class TimeSpan(_TimePrimitive):
    """Represents an extent in time bounded by begin and end dateTimes."""

    __name__ = "TimeSpan"

    def __init__(
        self,
        ns: Optional[str] = None,
        id: Optional[str] = None,
        target_id: Optional[str] = None,
        begin: Optional[KmlDateTime] = None,
        end: Optional[KmlDateTime] = None,
    ) -> None:
        super().__init__(ns=ns, id=id, target_id=target_id)
        self.begin = begin
        self.end = end

    def from_element(self, element: Element) -> None:
        super().from_element(element)
        begin = element.find(f"{self.ns}begin")
        if begin is not None:
            self.begin = KmlDateTime.parse(begin.text)
        end = element.find(f"{self.ns}end")
        if end is not None:
            self.end = KmlDateTime.parse(end.text)

    def etree_element(self) -> Element:
        element = super().etree_element()
        if self.begin is not None:
            text = str(self.begin)
            if text:
                begin = config.etree.SubElement(  # type: ignore[attr-defined]
                    element,
                    f"{self.ns}begin",
                )
                begin.text = text
        if self.end is not None:
            text = str(self.end)
            if text:
                end = config.etree.SubElement(  # type: ignore[attr-defined]
                    element,
                    f"{self.ns}end",
                )
                end.text = text
        if self.begin == self.end is None:
            raise ValueError("Either begin, end or both must be set")
        # TODO test if end > begin
        return element



# Copyright (C) 2012 - 2020  Christian Ledermann
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA

"""Abstract base classes"""
import logging
from typing import Optional
from typing import Tuple
from typing import cast

from fastkml import config
from fastkml.helpers import o_from_attr
from fastkml.helpers import o_to_attr
from fastkml.types import Element
from fastkml.types import KmlObjectMap

logger = logging.getLogger(__name__)


class _XMLObject:
    """XML Baseclass."""

    __name__ = ""
    kml_object_mapping: Tuple[KmlObjectMap, ...] = ()

    def __init__(self, ns: Optional[str] = None) -> None:
        """Initialize the XML Object."""
        self.ns: str = config.KMLNS if ns is None else ns

    def etree_element(self) -> Element:
        """Return the KML Object as an Element."""
        if self.__name__:
            element: Element = config.etree.Element(  # type: ignore[attr-defined]
                f"{self.ns}{self.__name__}"
            )
        else:
            raise NotImplementedError(
                "Call of abstract base class, subclasses implement this!"
            )
        for mapping in self.kml_object_mapping:
            mapping["to_kml"](self, element, **mapping)
        return element

    def from_element(self, element: Element) -> None:
        """Load the KML Object from an Element."""
        if f"{self.ns}{self.__name__}" != element.tag:
            raise TypeError("Call of abstract base class, subclasses implement this!")
        for mapping in self.kml_object_mapping:
            mapping["from_kml"](self, element, **mapping)

    def from_string(self, xml_string: str) -> None:
        """Load the KML Object from serialized xml."""
        self.from_element(
            cast(Element, config.etree.XML(xml_string))  # type: ignore[attr-defined]
        )

    def to_string(self, prettyprint: bool = True) -> str:
        """Return the KML Object as serialized xml."""
        try:
            return cast(
                str,
                config.etree.tostring(  # type: ignore[attr-defined]
                    self.etree_element(),
                    encoding="UTF-8",
                    pretty_print=prettyprint,
                ).decode("UTF-8"),
            )
        except TypeError:
            return cast(
                str,
                config.etree.tostring(  # type: ignore[attr-defined]
                    self.etree_element(), encoding="UTF-8"
                ).decode("UTF-8"),
            )


class _BaseObject(_XMLObject):
    """
    Base class for all KML objects.

    This is an abstract base class and cannot be used directly in a
    KML file. It provides the id attribute, which allows unique
    identification of a KML element, and the targetId attribute,
    which is used to reference objects that have already been loaded into
    Google Earth. The id attribute must be assigned if the <Update>
    mechanism is to be used.
    """

    id = None
    target_id = None
    kml_object_mapping: Tuple[KmlObjectMap, ...] = (
        {
            "kml_attr": "id",
            "obj_attr": "id",
            "from_kml": o_from_attr,
            "to_kml": o_to_attr,
            "required": False,
            "validator": None,
        },
        {
            "kml_attr": "targetId",
            "obj_attr": "target_id",
            "from_kml": o_from_attr,
            "to_kml": o_to_attr,
            "required": False,
            "validator": None,
        },
    )

    def __init__(
        self,
        ns: Optional[str] = None,
        id: Optional[str] = None,
        target_id: Optional[str] = None,
    ) -> None:
        """Initialize the KML Object."""
        super().__init__(ns)
        self.id = id
        self.target_id = target_id

    def etree_element(self) -> Element:
        """Return the KML Object as an Element."""
        return super().etree_element()

    def from_element(self, element: Element) -> None:
        """Load the KML Object from an Element."""
        super().from_element(element)



# Copyright (C) 2021 - 2022  Christian Ledermann
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA

"""Test the styles classes."""

from fastkml import styles
from tests.base import Lxml
from tests.base import StdLibrary


class TestStdLibrary(StdLibrary):
    """Test with the standard library."""

    def test_style_url(self) -> None:
        url = styles.StyleUrl(id="id-0", url="#style-0", target_id="target-0")

        serialized = url.to_string()

        assert '<kml:styleUrl xmlns:kml="http://www.opengis.net/kml/2.2"' in serialized
        assert 'id="id-0"' in serialized
        assert 'targetId="target-0"' in serialized
        assert ">#style-0</kml:styleUrl>" in serialized

    def test_style_url_read(self) -> None:
        url = styles.StyleUrl()

        url.from_string(
            '<kml:styleUrl xmlns:kml="http://www.opengis.net/kml/2.2"'
            ' id="id-0" targetId="target-0">#style-0</kml:styleUrl>'
        )

        assert url.id == "id-0"
        assert url.url == "#style-0"
        assert url.target_id == "target-0"

    def test_icon_style(self) -> None:
        icons = styles.IconStyle(
            id="id-0",
            target_id="target-0",
            color="ff0000ff",
            color_mode="random",
            scale=1.0,
            heading=0,
            icon_href="http://example.com/icon.png",
        )

        serialized = icons.to_string()

        assert '<kml:IconStyle xmlns:kml="http://www.opengis.net/kml/2.2"' in serialized
        assert 'id="id-0"' in serialized
        assert 'targetId="target-0"' in serialized
        assert "<kml:color>ff0000ff</kml:color>" in serialized
        assert "<kml:colorMode>random</kml:colorMode>" in serialized
        assert "<kml:scale>1.0</kml:scale>" in serialized
        assert "<kml:heading>0</kml:heading>" in serialized
        assert "<kml:Icon>" in serialized
        assert "<kml:href>http://example.com/icon.png</kml:href>" in serialized
        assert "</kml:Icon>" in serialized
        assert "</kml:IconStyle>" in serialized

    def test_icon_style_read(self) -> None:
        icons = styles.IconStyle()

        icons.from_string(
            '<kml:IconStyle xmlns:kml="http://www.opengis.net/kml/2.2" '
            'id="id-1" targetId="target-1">'
            "<kml:color>ff2200ff</kml:color><kml:colorMode>random</kml:colorMode>"
            "<kml:scale>5</kml:scale><kml:heading>20</kml:heading><kml:Icon>"
            "<kml:href>http://example.com/icon.png</kml:href></kml:Icon>"
            "</kml:IconStyle>"
        )

        assert icons.id == "id-1"
        assert icons.target_id == "target-1"
        assert icons.color == "ff2200ff"
        assert icons.color_mode == "random"
        assert icons.scale == 5.0
        assert icons.icon_href == "http://example.com/icon.png"
        assert icons.heading == 20.0

    def test_line_style(self) -> None:

        lines = styles.LineStyle(
            id="id-0",
            target_id="target-0",
            color="ff0000ff",
            color_mode="normal",
            width=1.0,
        )

        serialized = lines.to_string()

        assert '<kml:LineStyle xmlns:kml="http://www.opengis.net/kml/2.2"' in serialized
        assert 'id="id-0"' in serialized
        assert 'targetId="target-0"' in serialized
        assert "<kml:color>ff0000ff</kml:color>" in serialized
        assert "<kml:colorMode>normal</kml:colorMode>" in serialized
        assert "<kml:width>1.0</kml:width>" in serialized
        assert "</kml:LineStyle>" in serialized

    def test_line_style_read(self) -> None:
        lines = styles.LineStyle()

        lines.from_string(
            '<kml:LineStyle xmlns:kml="http://www.opengis.net/kml/2.2" '
            'id="id-l0" targetId="target-line">\n'
            "  <kml:color>ffaa00ff</kml:color>\n"
            "  <kml:colorMode>normal</kml:colorMode>\n"
            "  <kml:width>3.0</kml:width>\n"
            "</kml:LineStyle>\n"
        )

        assert lines.id == "id-l0"
        assert lines.target_id == "target-line"
        assert lines.color == "ffaa00ff"
        assert lines.color_mode == "normal"
        assert lines.width == 3.0

    def test_poly_style(self) -> None:
        ps = styles.PolyStyle(
            id="id-0",
            target_id="target-0",
            color="ff0000ff",
            color_mode="random",
            fill=0,
            outline=1,
        )

        serialized = ps.to_string()

        assert '<kml:PolyStyle xmlns:kml="http://www.opengis.net/kml/2.2"' in serialized
        assert 'id="id-0"' in serialized
        assert 'targetId="target-0"' in serialized
        assert "<kml:color>ff0000ff</kml:color>" in serialized
        assert "<kml:colorMode>random</kml:colorMode>" in serialized
        assert "<kml:fill>0</kml:fill>" in serialized
        assert "<kml:outline>1</kml:outline>" in serialized
        assert "</kml:PolyStyle>" in serialized

    def test_poly_style_read(self) -> None:
        ps = styles.PolyStyle()

        ps.from_string(
            '<kml:PolyStyle xmlns:kml="http://www.opengis.net/kml/2.2" '
            'id="id-1" targetId="target-1">'
            "<kml:color>ffaabbff</kml:color>"
            "<kml:colorMode>normal</kml:colorMode>"
            "<kml:fill>1</kml:fill>"
            "<kml:outline>0</kml:outline>"
            "</kml:PolyStyle>"
        )

        assert ps.id == "id-1"
        assert ps.target_id == "target-1"
        assert ps.color == "ffaabbff"
        assert ps.color_mode == "normal"
        assert ps.fill == 1
        assert ps.outline == 0

    def test_label_style(self) -> None:
        ls = styles.LabelStyle(
            id="id-0",
            target_id="target-0",
            color="ff0000ff",
            color_mode="random",
            scale=1.0,
        )

        serialized = ls.to_string()

        assert (
            '<kml:LabelStyle xmlns:kml="http://www.opengis.net/kml/2.2"' in serialized
        )
        assert 'id="id-0"' in serialized
        assert 'targetId="target-0"' in serialized
        assert "<kml:color>ff0000ff</kml:color>" in serialized
        assert "<kml:colorMode>random</kml:colorMode>" in serialized
        assert "<kml:scale>1.0</kml:scale>" in serialized
        assert "</kml:LabelStyle>" in serialized

    def test_label_style_read(self) -> None:
        ls = styles.LabelStyle()

        ls.from_string(
            '<kml:LabelStyle xmlns:kml="http://www.opengis.net/kml/2.2" '
            'id="id-1" targetId="target-1">'
            "<kml:color>ff001122</kml:color>"
            "<kml:colorMode>normal</kml:colorMode>"
            "<kml:scale>2.2</kml:scale>"
            "</kml:LabelStyle>"
        )

        assert ls.id == "id-1"
        assert ls.target_id == "target-1"
        assert ls.color == "ff001122"
        assert ls.color_mode == "normal"
        assert ls.scale == 2.2

    def test_balloon_style(self) -> None:
        bs = styles.BalloonStyle(
            id="id-0",
            target_id="target-0",
            bg_color="7fff0000",
            text_color="ff00ff00",
            text="<b>Hello</b>",
            display_mode="hide",
        )

        serialized = bs.to_string()

        assert (
            '<kml:BalloonStyle xmlns:kml="http://www.opengis.net/kml/2.2"' in serialized
        )
        assert 'id="id-0"' in serialized
        assert 'targetId="target-0">' in serialized
        assert "<kml:bgColor>7fff0000</kml:bgColor>" in serialized
        assert "<kml:textColor>ff00ff00</kml:textColor>" in serialized
        assert "<kml:text>&lt;b&gt;Hello&lt;/b&gt;</kml:text>" in serialized
        assert "<kml:displayMode>hide</kml:displayMode>" in serialized
        assert "</kml:BalloonStyle>" in serialized

    def test_balloon_style_read(self) -> None:
        bs = styles.BalloonStyle()

        bs.from_string(
            '<kml:BalloonStyle xmlns:kml="http://www.opengis.net/kml/2.2" '
            'id="id-7" targetId="target-6">'
            "<kml:bgColor>7fff1144</kml:bgColor>"
            "<kml:textColor>ff11ff22</kml:textColor>"
            "<kml:text>&lt;b&gt;World&lt;/b&gt;</kml:text>"
            "<kml:displayMode>default</kml:displayMode>"
            "</kml:BalloonStyle>"
        )

        assert bs.id == "id-7"
        assert bs.target_id == "target-6"
        assert bs.bg_color == "7fff1144"
        assert bs.text_color == "ff11ff22"
        assert bs.text == "<b>World</b>"
        assert bs.display_mode == "default"

    def test_style(self) -> None:
        icons = styles.IconStyle(
            id="id-i0",
            target_id="target-i0",
            color="ff0000ff",
            color_mode="random",
            scale=1.0,
            heading=0,
            icon_href="http://example.com/icon.png",
        )
        lines = styles.LineStyle(
            id="id-l0",
            target_id="target-l0",
            color="ff0000ff",
            color_mode="normal",
            width=1.0,
        )
        bs = styles.BalloonStyle(
            id="id-b0",
            target_id="target-b0",
            bg_color="7fff0000",
            text_color="ff00ff00",
            text="<b>Hello</b>",
            display_mode="hide",
        )
        ls = styles.LabelStyle(
            id="id-a0",
            target_id="target-a0",
            color="ff0000ff",
            color_mode="random",
            scale=1.0,
        )
        ps = styles.PolyStyle(
            id="id-p0",
            target_id="target-p0",
            color="ff0000ff",
            color_mode="random",
            fill=0,
            outline=1,
        )
        style = styles.Style(
            id="id-0",
            target_id="target-0",
            styles=[icons, lines, bs, ls, ps],
        )

        serialized = style.to_string()

        assert '<kml:Style xmlns:kml="http://www.opengis.net/kml/2.2" '
        assert 'id="id-0" '
        assert 'targetId="target-0">'
        assert "<kml:IconStyle" in serialized
        assert "<kml:Icon>" in serialized
        assert "</kml:Icon>" in serialized
        assert "</kml:IconStyle>"
        assert "<kml:LineStyle" in serialized
        assert "<kml:BalloonStyle " in serialized
        assert "<kml:LabelStyle " in serialized
        assert "<kml:PolyStyle " in serialized
        assert "</kml:PolyStyle>" in serialized
        assert "</kml:Style>" in serialized

    def test_style_read(self) -> None:
        style = styles.Style()

        style.from_string(
            '<kml:Style xmlns:kml="http://www.opengis.net/kml/2.2" id="id-0" targetId="target-0">'
            '<kml:IconStyle id="id-i0" targetId="target-i0">'
            "<kml:color>ff0000ff</kml:color>"
            "<kml:colorMode>random</kml:colorMode>"
            "<kml:scale>1.0</kml:scale>"
            "<kml:heading>0</kml:heading>"
            "<kml:Icon>"
            "<kml:href>http://example.com/icon.png</kml:href>"
            "</kml:Icon>"
            "</kml:IconStyle>"
            '<kml:LineStyle id="id-l0" targetId="target-l0">'
            "<kml:color>ff0000ff</kml:color>"
            "<kml:colorMode>normal</kml:colorMode>"
            "<kml:width>1.0</kml:width>"
            "</kml:LineStyle>"
            '<kml:BalloonStyle id="id-b0" targetId="target-b0">'
            "<kml:bgColor>7fff0000</kml:bgColor>"
            "<kml:textColor>ff00ff00</kml:textColor>"
            "<kml:text>&lt;b&gt;Hello&lt;/b&gt;</kml:text>"
            "<kml:displayMode>hide</kml:displayMode>"
            "</kml:BalloonStyle>"
            '<kml:LabelStyle id="id-a0" targetId="target-a0">'
            "<kml:color>ff0000ff</kml:color>"
            "<kml:colorMode>random</kml:colorMode>"
            "<kml:scale>1.0</kml:scale>"
            "</kml:LabelStyle>"
            '<kml:PolyStyle id="id-p0" targetId="target-p0">'
            "<kml:color>ff0000ff</kml:color>"
            "<kml:colorMode>random</kml:colorMode>"
            "<kml:fill>0</kml:fill>"
            "<kml:outline>1</kml:outline>"
            "</kml:PolyStyle>"
            "</kml:Style>"
        )

        assert style.id == "id-0"
        assert style.target_id == "target-0"
        assert list(style.styles())[0].id == "id-b0"
        assert list(style.styles())[0].target_id == "target-b0"
        assert list(style.styles())[0].bg_color == "7fff0000"  # type: ignore[union-attr]
        assert list(style.styles())[0].text_color == "ff00ff00"  # type: ignore[union-attr]
        assert list(style.styles())[0].text == "<b>Hello</b>"  # type: ignore[union-attr]
        assert list(style.styles())[0].display_mode == "hide"  # type: ignore[union-attr]

        assert list(style.styles())[1].id == "id-i0"
        assert list(style.styles())[1].target_id == "target-i0"
        assert list(style.styles())[1].color == "ff0000ff"  # type: ignore[union-attr]
        assert list(style.styles())[1].color_mode == "random"  # type: ignore[union-attr]
        assert list(style.styles())[1].scale == 1.0  # type: ignore[union-attr]
        assert list(style.styles())[1].heading == 0  # type: ignore[union-attr]
        assert list(style.styles())[1].icon_href == "http://example.com/icon.png"  # type: ignore[union-attr]

        assert list(style.styles())[2].id == "id-a0"
        assert list(style.styles())[2].target_id == "target-a0"
        assert list(style.styles())[2].color == "ff0000ff"  # type: ignore[union-attr]
        assert list(style.styles())[2].color_mode == "random"  # type: ignore[union-attr]
        assert list(style.styles())[2].scale == 1.0  # type: ignore[union-attr]

        assert list(style.styles())[3].id == "id-l0"
        assert list(style.styles())[3].target_id == "target-l0"
        assert list(style.styles())[3].color == "ff0000ff"  # type: ignore[union-attr]
        assert list(style.styles())[3].color_mode == "normal"  # type: ignore[union-attr]
        assert list(style.styles())[3].width == 1.0  # type: ignore[union-attr]

        assert list(style.styles())[4].id == "id-p0"
        assert list(style.styles())[4].target_id == "target-p0"
        assert list(style.styles())[4].color == "ff0000ff"  # type: ignore[union-attr]
        assert list(style.styles())[4].color_mode == "random"  # type: ignore[union-attr]
        assert list(style.styles())[4].fill == 0  # type: ignore[union-attr]
        assert list(style.styles())[4].outline == 1  # type: ignore[union-attr]

    def test_stylemap(self) -> None:
        url = styles.StyleUrl(id="id-0", url="#style-0", target_id="target-0")
        icons = styles.IconStyle(
            id="id-i0",
            target_id="target-i0",
            color="ff0000ff",
            color_mode="random",
            scale=1.0,
            heading=0,
            icon_href="http://example.com/icon.png",
        )
        lines = styles.LineStyle(
            id="id-l0",
            target_id="target-l0",
            color="ff0000ff",
            color_mode="normal",
            width=1.0,
        )
        bs = styles.BalloonStyle(
            id="id-b0",
            target_id="target-b0",
            bg_color="7fff0000",
            text_color="ff00ff00",
            text="<b>Hello</b>",
            display_mode="hide",
        )
        ls = styles.LabelStyle(
            id="id-a0",
            target_id="target-a0",
            color="ff0000ff",
            color_mode="random",
            scale=1.0,
        )
        ps = styles.PolyStyle(
            id="id-p0",
            target_id="target-p0",
            color="ff0000ff",
            color_mode="random",
            fill=0,
            outline=1,
        )
        style = styles.Style(
            id="id-0",
            target_id="target-0",
            styles=[icons, lines, bs, ls, ps],
        )
        sm = styles.StyleMap(
            id="id-sm-0",
            target_id="target-sm-0",
            normal=url,
            highlight=style,
        )

        serialized = sm.to_string()

        assert '<kml:StyleMap xmlns:kml="http://www.opengis.net/kml/2.2" ' in serialized
        assert 'id="id-sm-0"' in serialized
        assert 'targetId="target-sm-0"' in serialized
        assert "<kml:Pair>" in serialized
        assert "<kml:key>normal</kml:key>" in serialized
        assert "<kml:styleUrl " in serialized
        assert "</kml:Pair>" in serialized
        assert "<kml:Pair>" in serialized
        assert "<kml:key>highlight</kml:key>" in serialized
        assert "<kml:Style " in serialized
        assert "<kml:IconStyle " in serialized
        assert "<kml:color>ff0000ff</kml:color>" in serialized
        assert "<kml:colorMode>random</kml:colorMode>" in serialized
        assert "<kml:scale>1.0</kml:scale>" in serialized
        assert "<kml:heading>0</kml:heading>" in serialized
        assert "<kml:Icon>" in serialized
        assert "<kml:href>http://example.com/icon.png</kml:href>" in serialized
        assert "</kml:Icon>" in serialized
        assert "</kml:IconStyle>" in serialized
        assert "<kml:LineStyle " in serialized
        assert "<kml:color>ff0000ff</kml:color>" in serialized
        assert "<kml:colorMode>normal</kml:colorMode>" in serialized
        assert "<kml:width>1.0</kml:width>" in serialized
        assert "</kml:LineStyle>" in serialized
        assert "<kml:BalloonStyle " in serialized
        assert "<kml:bgColor>7fff0000</kml:bgColor>" in serialized
        assert "<kml:textColor>ff00ff00</kml:textColor>" in serialized
        assert "<kml:text>&lt;b&gt;Hello&lt;/b&gt;</kml:text>" in serialized
        assert "<kml:displayMode>hide</kml:displayMode>" in serialized
        assert "</kml:BalloonStyle>" in serialized
        assert "<kml:LabelStyle " in serialized
        assert "<kml:color>ff0000ff</kml:color>" in serialized
        assert "<kml:colorMode>random</kml:colorMode>" in serialized
        assert "<kml:scale>1.0</kml:scale>" in serialized
        assert "</kml:LabelStyle>" in serialized
        assert "<kml:PolyStyle " in serialized
        assert "<kml:color>ff0000ff</kml:color>" in serialized
        assert "<kml:colorMode>random</kml:colorMode>" in serialized
        assert "<kml:fill>0</kml:fill>" in serialized
        assert "<kml:outline>1</kml:outline>" in serialized
        assert "</kml:PolyStyle>" in serialized
        assert "</kml:Style>" in serialized
        assert "</kml:Pair>" in serialized
        assert "</kml:StyleMap>" in serialized

    def test_stylemap_read(self) -> None:
        sm = styles.StyleMap()
        sm.from_string(
            """
            <kml:StyleMap xmlns:kml="http://www.opengis.net/kml/2.2"
            id="id-sm-0" targetId="target-sm-0">
            <kml:Pair>
                <kml:key>highlight</kml:key>
                <kml:styleUrl id="id-u0" targetId="target-u0">#style-0</kml:styleUrl>
            </kml:Pair>
            <kml:Pair>
                <kml:key>normal</kml:key>
                <kml:Style id="id-0" targetId="target-0">
                    <kml:IconStyle id="id-i0" targetId="target-i0">
                        <kml:color>ff0000ff</kml:color>
                        <kml:colorMode>random</kml:colorMode>
                        <kml:scale>1.0</kml:scale>
                        <kml:heading>0</kml:heading>
                        <kml:Icon>
                            <kml:href>http://example.com/icon.png</kml:href>
                        </kml:Icon>
                    </kml:IconStyle>
                    <kml:LineStyle id="id-l0" targetId="target-l0">
                        <kml:color>ff0000ff</kml:color>
                        <kml:colorMode>normal</kml:colorMode>
                        <kml:width>1.0</kml:width>
                    </kml:LineStyle>
                    <kml:BalloonStyle id="id-b0" targetId="target-b0">
                        <kml:bgColor>7fff0000</kml:bgColor>
                        <kml:textColor>ff00ff00</kml:textColor>
                        <kml:text>&lt;b&gt;Hello&lt;/b&gt;</kml:text>
                        <kml:displayMode>hide</kml:displayMode>
                    </kml:BalloonStyle>
                    <kml:LabelStyle id="id-a0" targetId="target-a0">
                        <kml:color>ff0000ff</kml:color>
                        <kml:colorMode>random</kml:colorMode>
                        <kml:scale>1.0</kml:scale>
                    </kml:LabelStyle>
                    <kml:PolyStyle id="id-p0" targetId="target-p0">
                        <kml:color>ff0000ff</kml:color>
                        <kml:colorMode>random</kml:colorMode>
                        <kml:fill>0</kml:fill>
                        <kml:outline>1</kml:outline>
                    </kml:PolyStyle>
                </kml:Style>
            </kml:Pair>
            </kml:StyleMap>
        """
        )

        assert sm.id == "id-sm-0"
        assert sm.target_id == "target-sm-0"
        assert sm.normal.id == "id-0"  # type: ignore[union-attr]
        assert sm.normal.target_id == "target-0"  # type: ignore[union-attr]
        assert sm.highlight.id == "id-u0"  # type: ignore[union-attr]
        assert sm.highlight.target_id == "target-u0"  # type: ignore[union-attr]


class TestLxml(Lxml, TestStdLibrary):
    """Test with lxml."""



# Copyright (C) 2012 - 2022 Christian Ledermann
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA

"""
With the launch of Google Earth 5.0, Google has provided extensions to KML
to support a number of new features. These extensions use the gx prefix
and the following namespace URI::

    xmlns:gx="http://www.google.com/kml/ext/2.2"

This namespace URI must be added to the <kml> element in any KML file
using gx-prefixed elements::

    <kml
        xmlns="http://www.opengis.net/kml/2.2"
        xmlns:gx="http://www.google.com/kml/ext/2.2"
    >

Extensions to KML may not be supported in all geo-browsers. If your
browser doesn't support particular extensions, the data in those
extensions should be silently ignored, and the rest of the KML file
should load without errors.

Elements that currently use the gx prefix are:

* gx:altitudeMode
* gx:altitudeOffset
* gx:angles
* gx:AnimatedUpdate
* gx:balloonVisibility
* gx:coord
* gx:delayedStart
* gx:drawOrder
* gx:duration
* gx:FlyTo
* gx:flyToMode
* gx:h
* gx:horizFov
* gx:interpolate
* gx:labelVisibility
* gx:LatLonQuad
* gx:MultiTrack
* gx:vieweroptions
* gx:outerColor
* gx:outerWidth
* gx:physicalWidth
* gx:Playlist
* gx:playMode
* gx:SoundCue
* gx:TimeSpan
* gx:TimeStamp
* gx:Tour
* gx:TourControl
* gx:TourPrimitive
* gx:Track
* gx:ViewerOptions
* gx:w
* gx:Wait
* gx:x
* gx:y

The complete XML schema for elements in this extension namespace is
located at http://developers.google.com/kml/schema/kml22gx.xsd.
"""

import logging
from typing import List
from typing import Optional
from typing import Union
from typing import cast

from pygeoif.geometry import GeometryCollection
from pygeoif.geometry import LineString
from pygeoif.geometry import MultiLineString
from pygeoif.types import PointType

from fastkml.config import GXNS as NS
from fastkml.geometry import Geometry
from fastkml.types import Element

logger = logging.getLogger(__name__)


class GxGeometry(Geometry):
    def __init__(
        self,
        ns: None = None,
        id: None = None,
    ) -> None:
        """
        gxgeometry: a read-only subclass of geometry supporting gx: features,
        like gx:Track
        """
        super().__init__(ns, id)
        self.ns = NS if ns is None else ns

    def _get_geometry(self, element: Element) -> Optional[LineString]:
        # Track
        if element.tag == (f"{self.ns}Track"):
            coords = self._get_coordinates(element)
            self._get_geometry_spec(element)
            return LineString(
                coords,
            )
        return None

    def _get_multigeometry(
        self,
        element: Element,
    ) -> Union[MultiLineString, GeometryCollection, None]:
        # MultiTrack
        geoms = []
        if element.tag == (f"{self.ns}MultiTrack"):
            tracks = element.findall(f"{self.ns}Track")
            for track in tracks:
                self._get_geometry_spec(track)
                geoms.append(
                    LineString(
                        self._get_coordinates(track),
                    )
                )

        geom_types = {geom.geom_type for geom in geoms}
        if len(geom_types) > 1:
            return GeometryCollection(geoms)
        if "LineString" in geom_types:
            return MultiLineString.from_linestrings(*geoms)
        return None

    def _get_coordinates(self, element: Element) -> List[PointType]:
        coordinates = element.findall(f"{self.ns}coord")
        if coordinates is not None:
            return [
                cast(PointType, tuple(float(c) for c in coord.text.strip().split()))
                for coord in coordinates
            ]
        return []  # type: ignore[unreachable]


__all__ = ["GxGeometry"]



# Copyright (C) 2022  Christian Ledermann
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA

"""Test the (Abstract)Views classes."""

import datetime

from fastkml import times
from fastkml import views
from tests.base import Lxml
from tests.base import StdLibrary


class TestStdLibrary(StdLibrary):
    """Test with the standard library."""

    def test_create_camera(self) -> None:
        """Test the creation of a camera."""
        time_span = times.TimeSpan(
            id="time-span-id",
            begin=times.KmlDateTime(datetime.datetime(2019, 1, 1)),
            end=times.KmlDateTime(datetime.datetime(2019, 1, 2)),
        )

        camera = views.Camera(
            id="cam-id",
            target_id="target-cam-id",
            heading=10,
            tilt=20,
            roll=30,
            altitude=40,
            altitude_mode="relativeToGround",
            latitude=50,
            longitude=60,
            time_primitive=time_span,
        )

        assert camera.heading == 10
        assert camera.tilt == 20
        assert camera.roll == 30
        assert camera.altitude == 40
        assert camera.altitude_mode == "relativeToGround"
        assert camera.latitude == 50
        assert camera.longitude == 60
        assert camera.id == "cam-id"
        assert camera.target_id == "target-cam-id"
        assert camera.begin == times.KmlDateTime(datetime.datetime(2019, 1, 1))
        assert camera.end == times.KmlDateTime(datetime.datetime(2019, 1, 2))
        assert camera.to_string()

    def test_camera_read(self) -> None:
        """Test the reading of a camera."""
        camera_xml = (
            '<kml:Camera xmlns:kml="http://www.opengis.net/kml/2.2" '
            'id="cam-id" targetId="target-cam-id">'
            '<kml:TimeSpan id="time-span-id">'
            "<kml:begin>2019-01-01T00:00:00</kml:begin>"
            "<kml:end>2019-01-02T00:00:00</kml:end>"
            "</kml:TimeSpan>"
            "<kml:longitude>60</kml:longitude>"
            "<kml:latitude>50</kml:latitude>"
            "<kml:altitude>40</kml:altitude>"
            "<kml:heading>10</kml:heading>"
            "<kml:tilt>20</kml:tilt>"
            "<kml:roll>30</kml:roll>"
            "<kml:altitudeMode>relativeToGround</kml:altitudeMode>"
            "</kml:Camera>"
        )
        camera = views.Camera()

        camera.from_string(camera_xml)

        assert camera.heading == 10
        assert camera.tilt == 20
        assert camera.roll == 30
        assert camera.altitude == 40
        assert camera.altitude_mode == "relativeToGround"
        assert camera.latitude == 50
        assert camera.longitude == 60
        assert camera.id == "cam-id"
        assert camera.target_id == "target-cam-id"
        assert camera.begin == times.KmlDateTime(datetime.datetime(2019, 1, 1))
        assert camera.end == times.KmlDateTime(datetime.datetime(2019, 1, 2))

    def test_create_look_at(self) -> None:
        time_stamp = times.TimeStamp(
            id="time-span-id",
            timestamp=times.KmlDateTime(datetime.datetime(2019, 1, 1)),
        )

        look_at = views.LookAt(
            id="look-at-id",
            target_id="target-look-at-id",
            heading=10,
            tilt=20,
            range=30,
            altitude_mode="relativeToGround",
            latitude=50,
            longitude=60,
            time_primitive=time_stamp,
        )

        assert look_at.heading == 10
        assert look_at.tilt == 20
        assert look_at.range == 30
        assert look_at.altitude_mode == "relativeToGround"
        assert look_at.latitude == 50
        assert look_at.longitude == 60
        assert look_at.id == "look-at-id"
        assert look_at.target_id == "target-look-at-id"
        assert look_at._timestamp.timestamp.dt == datetime.datetime(2019, 1, 1)
        assert look_at.begin is None
        assert look_at.end is None
        assert look_at.to_string()

    def test_look_at_read(self) -> None:
        look_at_xml = (
            '<kml:LookAt xmlns:kml="http://www.opengis.net/kml/2.2" id="look-at-id" '
            'targetId="target-look-at-id">'
            "<kml:longitude>60</kml:longitude>"
            "<kml:latitude>50</kml:latitude>"
            "<kml:heading>10</kml:heading>"
            "<kml:tilt>20</kml:tilt>"
            "<kml:altitudeMode>relativeToGround</kml:altitudeMode>"
            '<kml:TimeStamp id="time-span-id">'
            "<kml:when>2019-01-01T00:00:00</kml:when>"
            "</kml:TimeStamp>"
            "<kml:range>30</kml:range>"
            "</kml:LookAt>"
        )
        look_at = views.LookAt()

        look_at.from_string(look_at_xml)

        assert look_at.heading == 10
        assert look_at.tilt == 20
        assert look_at.range == 30
        assert look_at.altitude_mode == "relativeToGround"
        assert look_at.latitude == 50
        assert look_at.longitude == 60
        assert look_at.id == "look-at-id"
        assert look_at.target_id == "target-look-at-id"
        assert look_at._timestamp.timestamp.dt == datetime.datetime(2019, 1, 1)
        assert look_at.begin is None
        assert look_at.end is None


class TestLxml(Lxml, TestStdLibrary):
    """Test with lxml."""



import os
import sys

from setuptools import find_packages
from setuptools import setup
from setuptools.command.test import test as TestCommand


class PyTest(TestCommand):
    def finalize_options(self) -> None:
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self) -> None:
        # import here, cause outside the eggs aren't loaded
        import pytest

        errno = pytest.main(self.test_args)
        sys.exit(errno)


setup(
    name="fastkml",
    version="1.0.alpha.4",
    description="Fast KML processing in python",
    long_description=(
        open("README.rst").read()
        + "\n"
        + open(os.path.join("docs", "HISTORY.txt")).read()
    ),
    long_description_content_type="text/x-rst",
    classifiers=[
        "Topic :: Scientific/Engineering :: GIS",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)",
        # "Development Status :: 5 - Production/Stable",
        "Development Status :: 3 - Alpha",
        "Operating System :: OS Independent",
    ],  # Get strings from http://pypi.python.org/pypi?%3Aaction=list_classifiers
    keywords="GIS KML Google Maps OpenLayers",
    author="Christian Ledermann",
    author_email="christian.ledermann@gmail.com",
    url="https://github.com/cleder/fastkml",
    license="LGPL",
    packages=find_packages(exclude=["ez_setup", "examples", "tests"]),
    include_package_data=True,
    zip_safe=False,
    tests_require=["pytest"],
    cmdclass={"test": PyTest},
    python_requires=">=3.7",
    install_requires=[
        # -*- Extra requirements: -*-
        "pygeoif>=1.0.0",
        "python-dateutil",
        "setuptools",
        "typing_extensions",
    ],
    entry_points="""
    # -*- Entry points: -*-
    """,
)



# Copyright (C) 2021  Christian Ledermann
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA

"""Test the gx classes."""
from tests.base import Lxml
from tests.base import StdLibrary


class TestStdLibrary(StdLibrary):
    """Test with the standard library."""


class TestLxml(Lxml, TestStdLibrary):
    """Test with lxml."""



# Copyright (C) 2012 - 2021  Christian Ledermann
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA

"""
KML 2.2 supports new elements for including data about the author and related
website in your KML file. This information is displayed in geo search results,
both in Earth browsers such as Google Earth, and in other applications such as
Google Maps. The ascription elements used in KML are as follows:

atom:author element - parent element for atom:name
atom:name element - the name of the author
atom:link element - contains the href attribute
href attribute - URL of the web page containing the KML/KMZ file

These elements are defined in the Atom Syndication Format. The complete
specification is found at http://atompub.org.

This library only implements a subset of Atom that is useful with KML
"""

import logging
import re
from typing import Optional
from typing import Tuple

from fastkml.base import _XMLObject
from fastkml.config import ATOMNS as NS
from fastkml.helpers import o_from_attr
from fastkml.helpers import o_from_subelement_text
from fastkml.helpers import o_int_from_attr
from fastkml.helpers import o_to_attr
from fastkml.helpers import o_to_subelement_text
from fastkml.types import Element
from fastkml.types import KmlObjectMap

logger = logging.getLogger(__name__)
regex = r"(^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$)"
email_match = re.compile(regex).match


def check_email(email: str) -> bool:
    """Check if the email address is valid."""
    return bool(email_match(email))


class Link(_XMLObject):
    """
    Identifies a related Web page. The type of relation is defined by
    the rel attribute. A feed is limited to one alternate per type and
    hreflang.
    <link> is patterned after html's link element. It has one required
    attribute, href, and five optional attributes: rel, type, hreflang,
    title, and length.
    """

    __name__ = "link"

    kml_object_mapping: Tuple[KmlObjectMap, ...] = (
        {
            "kml_attr": "href",
            "obj_attr": "href",
            "from_kml": o_from_attr,
            "to_kml": o_to_attr,
            "required": True,
            "validator": None,
        },
        {
            "kml_attr": "rel",
            "obj_attr": "rel",
            "from_kml": o_from_attr,
            "to_kml": o_to_attr,
            "required": False,
            "validator": None,
        },
        {
            "kml_attr": "type",
            "obj_attr": "type",
            "from_kml": o_from_attr,
            "to_kml": o_to_attr,
            "required": False,
            "validator": None,
        },
        {
            "kml_attr": "hreflang",
            "obj_attr": "hreflang",
            "from_kml": o_from_attr,
            "to_kml": o_to_attr,
            "required": False,
            "validator": None,
        },
        {
            "kml_attr": "title",
            "obj_attr": "title",
            "from_kml": o_from_attr,
            "to_kml": o_to_attr,
            "required": False,
            "validator": None,
        },
        {
            "kml_attr": "length",
            "obj_attr": "length",
            "from_kml": o_int_from_attr,
            "to_kml": o_to_attr,
            "required": False,
            "validator": None,
        },
    )

    href = None
    # href is the URI of the referenced resource

    rel = None
    # rel contains a single link relationship type.
    # It can be a full URI, or one of the following predefined values
    # (default=alternate):
    # alternate: an alternate representation
    # enclosure: a related resource which is potentially large in size
    # and might require special handling, for example an audio or video
    # recording.
    # related: an document related to the entry or feed.
    # self: the feed itself.
    # via: the source of the information provided in the entry.

    type = None
    # indicates the media type of the resource

    hreflang = None
    # indicates the language of the referenced resource

    title = None
    # human readable information about the link

    length = None
    # the length of the resource, in bytes

    def __init__(
        self,
        ns: Optional[str] = None,
        href: Optional[str] = None,
        rel: Optional[str] = None,
        type: Optional[str] = None,
        hreflang: Optional[str] = None,
        title: Optional[str] = None,
        length: Optional[int] = None,
    ) -> None:
        self.ns: str = NS if ns is None else ns
        self.href = href
        self.rel = rel
        self.type = type
        self.hreflang = hreflang
        self.title = title
        self.length = length

    def from_element(self, element: Element) -> None:
        super().from_element(element)

    def etree_element(self) -> Element:
        return super().etree_element()


class _Person(_XMLObject):
    """
    <author> and <contributor> describe a person, corporation, or similar
    entity. It has one required element, name, and two optional elements:
    uri, email.
    """

    __name__ = ""
    kml_object_mapping: Tuple[KmlObjectMap, ...] = (
        {
            "kml_attr": "name",
            "obj_attr": "name",
            "from_kml": o_from_subelement_text,
            "to_kml": o_to_subelement_text,
            "required": True,
            "validator": None,
        },
        {
            "kml_attr": "uri",
            "obj_attr": "uri",
            "from_kml": o_from_subelement_text,
            "to_kml": o_to_subelement_text,
            "required": False,
            "validator": None,
        },
        {
            "kml_attr": "email",
            "obj_attr": "email",
            "from_kml": o_from_subelement_text,
            "to_kml": o_to_subelement_text,
            "required": False,
            "validator": check_email,
        },
    )

    name: Optional[str] = None
    # conveys a human-readable name for the person.

    uri: Optional[str] = None
    # contains a home page for the person.

    email: Optional[str] = None
    # contains an email address for the person.

    def __init__(
        self,
        ns: Optional[str] = None,
        name: Optional[str] = None,
        uri: Optional[str] = None,
        email: Optional[str] = None,
    ) -> None:
        self.ns: str = NS if ns is None else ns
        self.name = name
        self.uri = uri
        self.email = email

    def etree_element(self) -> Element:
        return super().etree_element()

    def from_element(self, element: Element) -> None:
        super().from_element(element)


class Author(_Person):
    """
    Return the names one author of the feed/entry.

    A feed/entry may have multiple authors.
    """

    __name__ = "author"


class Contributor(_Person):
    """
    Return the names one contributor to the feed/entry.

    A feed/entry may have multiple contributor elements.
    """

    __name__ = "contributor"


__all__ = ["Author", "Contributor", "Link"]



# Copyright (C) 2021  Christian Ledermann
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA

"""Test the configuration options."""

import xml.etree.ElementTree

import pytest

try:
    import lxml

    LXML = True
except ImportError:
    LXML = False

from fastkml import config


def test_set_etree_implementation_xml() -> None:
    config.set_etree_implementation(xml.etree.ElementTree)

    assert config.etree.__name__ == "xml.etree.ElementTree"


@pytest.mark.skipif(not LXML, reason="lxml not installed")
def test_set_etree_implementation_lxml() -> None:
    config.set_etree_implementation(lxml.etree)

    assert config.etree.__name__ == "lxml.etree"


def test_register_namespaces() -> None:
    """Register namespaces for use in etree."""
    config.set_etree_implementation(xml.etree.ElementTree)
    ns = {
        "real_person": "http://people.example.com",
        "role": "http://characters.example.com",
    }

    config.register_namespaces(**ns)

    for k, v in ns.items():
        assert config.etree._namespace_map[v] == k


def test_default_registered_namespaces() -> None:
    assert config.DEFAULT_NAME_SPACES == {
        "kml": "http://www.opengis.net/kml/2.2",
        "atom": "http://www.w3.org/2005/Atom",
        "gx": "http://www.google.com/kml/ext/2.2",
    }


def test_set_default_namespaces() -> None:
    """Set the default namespaces."""
    config.set_etree_implementation(xml.etree.ElementTree)
    config.etree._namespace_map = {}

    config.set_default_namespaces()

    for k, v in config.DEFAULT_NAME_SPACES.items():
        assert config.etree._namespace_map[v] == k
